# Copyright (c) 2017 Riverbed Technology, Inc.
#
# This software is licensed under the terms and conditions of the MIT License
# accompanying the software ("License").  This software is distributed "AS IS"
# as set forth in the License.

from os.path import basename
from django import forms
from steelscript.appfwk.libs.fields import Function
from steelscript.appfwk.apps.pcapmgr.models import PcapDataFile
from steelscript.appfwk.apps.datasource.models import TableField
from steelscript.appfwk.apps.datasource.forms import IntegerIDChoiceField


class PcapFileForm(forms.ModelForm):
    class Meta:
        model = PcapDataFile
        fields = ('description', 'datafile', 'file_type',
                  'start_time', 'end_time', 'pkt_count',
                  'packet_bytes',)


class PcapFileListForm(forms.ModelForm):
    class Meta:
        model = PcapDataFile
        fields = ('description', 'datafile', 'file_type',
                  'start_time', 'end_time', 'pkt_count',
                  'packet_bytes',)


def pcap_selection_preprocess(form, field, field_kwargs, params):
    pcap_mgr_files = PcapDataFile.objects.order_by('id')
    if pcap_mgr_files:
        choices = [(p.id, basename(p.datafile.name)) for p in pcap_mgr_files]
    else:
        choices = [('', '<No PCAP Manager files available>')]

    field_kwargs['choices'] = choices


def fields_add_pcapmgr_selection(obj,
                                 keyword='pcapmgrfile',
                                 label='PCAPMgrFile'):
    field = TableField(keyword=keyword, label=label,
                       field_cls=IntegerIDChoiceField,
                       pre_process_func=Function(pcap_selection_preprocess))
    field.save()
    obj.fields.add(field)



# Copyright (c) 2015 Riverbed Technology, Inc.
#
# This software is licensed under the terms and conditions of the MIT License
# accompanying the software ("License").  This software is distributed "AS IS"
# as set forth in the License.


from django.contrib import admin

from steelscript.appfwk.apps.db.models import ExistingIntervals


class IntervalsAdmin(admin.ModelAdmin):
    list_display = ('namespace', 'sourcefile', 'table', 'criteria',
                    'existing_intervals', 'table_handle')

    def existing_intervals(self, obj):

        intervals = []
        for interval in obj.intervals:
            interval.localize_tz(obj.tzinfo)
            intervals.append(str(interval))
        return '<br>'.join(intervals)

    existing_intervals.allow_tags = True


admin.site.register(ExistingIntervals, IntervalsAdmin)



# Copyright (c) 2017 Riverbed Technology, Inc.
#
# This software is licensed under the terms and conditions of the MIT License
# accompanying the software ("License").  This software is distributed "AS IS"
# as set forth in the License.

import logging
from datetime import datetime
import magic
from django.forms import ValidationError
from django.db import models
from django.db.models.fields.files import FieldFile
from steelscript.appfwk.apps.pcapmgr.filestore import pcap_store

WARN_PCAP = True
try:
    from steelscript.packets.core.pcap import pcap_info
    WARN_PCAP = False
except ImportError:
    def pcap_info(file_handle):
        return {'first_timestamp': 0.0,
                'last_timestamp': 0.0,
                'total_packets': 0,
                'total_bytes': 0}

logger = logging.getLogger(__name__)


class PcapFileField(models.FileField):
    """ Implements a FileField with support for magic file type checking
        in the fields clean() method.
    """
    # The class to wrap instance attributes in. Accessing the file object off
    # the instance will always return an instance of attr_class.
    attr_class = FieldFile

    SUPPORTED_FILES = {'PCAP (libpcap format)': 'tcpdump capture file',
                       'PCAP Next Generation': 'pcap-ng capture file'}
    # Magic read length taken from magic.py examples:
    # https://github.com/ahupp/python-magic/blob/master/magic.py
    MAGIC_LEN = 1024
    SIG_LEN = 20

    def __init__(self, *args, **kwargs):
        self.magic_file_type = None
        super(PcapFileField, self).__init__(*args, **kwargs)

    def clean(self, value, model_instance):
        self.magic_file_type, file_type_name = self.get_magic_type(value.file)

        if self.magic_file_type is not None:
            return super(PcapFileField, self).clean(value, model_instance)
        else:
            raise ValidationError("File type '{0}' is not supported by this"
                                  " file manager.".format(file_type_name))

    @classmethod
    def get_magic_type(cls, file_object):
        file_object.seek(0)
        magic_code = magic.from_buffer(file_object.read(cls.MAGIC_LEN))
        file_object.seek(0)
        file_type_name = None
        for type_name, sig in cls.SUPPORTED_FILES.viewitems():
            if sig == magic_code[:cls.SIG_LEN]:
                file_type_name = type_name
                break

        return file_type_name, magic_code


class DataFileBase(models.Model):
    """ Base Data File model. Used as a base class by PCAPDataFile. Can
        be used as base class for other file store types.
    """
    description = models.CharField(max_length=255, blank=True)
    uploaded_at = models.DateTimeField(auto_now_add=True)
    file_type = models.CharField(max_length=255, blank=True)

    def __unicode__(self):
        return self.description


class PcapDataFile(DataFileBase):
    """ PCAP and PCAP NG supporting Data File model.
        Uses the PcapFileField and the PcapFileField.SUPPORTED_FILES in order
        to implement support for only the correct file types.
    """
    datafile = PcapFileField(storage=pcap_store,
                             upload_to=pcap_store.location)
    start_time = models.DateTimeField(default=datetime.now, blank=True)
    end_time = models.DateTimeField(default=datetime.now, blank=True)
    pkt_count = models.IntegerField(blank=True)
    packet_bytes = models.IntegerField(blank=True)

    SUPPORTED_FILES = PcapFileField.SUPPORTED_FILES

    def __unicode__(self):
        s = '{dfile}({f_type}) - {desc} (saved:{at})'
        return s.format(dfile=self.datafile,
                        desc=self.description,
                        at=self.uploaded_at,
                        f_type=self.file_type)

    def save(self, *args, **kwargs):
        """ Saved model instance to the database. 2 cases supported: 1st is
            UI created PcapDataFile instance with the datafile instance
            being a PcapFileField wrapped in a FieldFile. Second is a
            manually created PcapDataFile generated when a file system sync
            operation finds files not present in the DB."""

        # See attr_class variable for datafile instance for reason why
        # an instance variable is being checked for a class type variable.
        f_type = self.datafile.field.magic_file_type

        # This f_type variable will exist if this instance is a UI
        # created PCAPDataFile. But if this is being picked up via
        # a file system sync operation then it will not exist.
        if f_type:
            self.file_type = f_type
            datafile_clean = True
        else:
            # self.file_type was manually set during sync operation.
            # But the built in pseudo file object can't be used yet.
            datafile_clean = False

        if self.file_type in self.SUPPORTED_FILES:
            if datafile_clean:
                # UI created instance. Simply use the built in pseudo file type
                # object.
                pinfo = pcap_info(self.datafile.file)
            else:
                # Manually created during sync. Have to open as a true python
                # file object.
                with open(self.datafile.file.name, 'rb') as f:
                    pinfo = pcap_info(f)

            # Assign the special pcap fields.
            self.start_time = \
                datetime.utcfromtimestamp(pinfo['first_timestamp'])
            self.end_time = datetime.utcfromtimestamp(pinfo['last_timestamp'])
            self.pkt_count = pinfo['total_packets']
            self.packet_bytes = pinfo['total_bytes']

        super(PcapDataFile, self).save(*args, **kwargs)



# Copyright (c) 2015 Riverbed Technology, Inc.
#
# This software is licensed under the terms and conditions of the MIT License
# accompanying the software ("License").  This software is distributed "AS IS"
# as set forth in the License.

from django.conf.urls import patterns, url

import steelscript.appfwk.apps.db.views as views


urlpatterns = patterns(
    'steelscript.appfwk.apps.db.views',

    url(r'^records/$',
        views.Records.as_view()),

    url(r'^handles/$',
        views.Handles.as_view()),
)



# Copyright (c) 2017 Riverbed Technology, Inc.
#
# This software is licensed under the terms and conditions of the MIT License
# accompanying the software ("License").  This software is distributed "AS IS"
# as set forth in the License.

from django.conf.urls import patterns, url
from rest_framework.urlpatterns import format_suffix_patterns

from steelscript.appfwk.apps.pcapmgr.views import PcapFileList, \
    PcapFileDetail, PcapFSSync, pcap_download


urlpatterns = patterns(
    '',

    url(r'^$',
        PcapFileList.as_view(),
        name='pcapfile-list'),

    url(r'^(?P<pcapfile_id>[0-9]+)/$',
        PcapFileDetail.as_view(),
        name='pcapfile-detail'),

    url(r'^add/$',
        PcapFileDetail.as_view(),
        name='pcapfile-add'),

    url(r'^sync/$',
        PcapFSSync.as_view(),
        name='pcapfs-sync'),

    url(r'^(?P<pcap_name>.*)/download/$',
        pcap_download,
        name='pcapfile-download'),

)

urlpatterns = format_suffix_patterns(urlpatterns)



# Copyright (c) 2015 Riverbed Technology, Inc.
#
# This software is licensed under the terms and conditions of the MIT License
# accompanying the software ("License").  This software is distributed "AS IS"
# as set forth in the License.


from django.conf import settings
from models import ExistingIntervals

if (not hasattr(settings, 'DB_SOLUTION')):
    raise Exception('settings.DB_SOLUTION not set')

elif (settings.DB_SOLUTION == 'elastic') or settings.TESTING:
    from steelscript.appfwk.apps.db.elastic import ElasticSearch, ColumnFilter
    storage = ElasticSearch()

else:
    raise Exception('Unrecognized settings.DB_SOLUTION: %s' %
                    settings.DB_SOLUTION)



# Copyright (c) 2015 Riverbed Technology, Inc.
#
# This software is licensed under the terms and conditions of the MIT License
# accompanying the software ("License").  This software is distributed "AS IS"
# as set forth in the License.

from django.db import models

from steelscript.appfwk.libs.fields import PickledObjectField


class ExistingIntervals(models.Model):
    """Store the existing time intervals in db for each table and
    a set of criteria fields (represented by the table_handle field).
    """
    namespace = models.CharField(max_length=20)

    sourcefile = models.CharField(max_length=200)

    table = models.CharField(max_length=50)

    criteria = PickledObjectField(null=True)

    table_handle = models.CharField(max_length=100, default="")

    intervals = PickledObjectField(null=True)

    tzinfo = PickledObjectField()



from os.path import basename
from django import template
from django.template.defaultfilters import stringfilter

register = template.Library()


@register.filter
@stringfilter
def filename(value):
    return basename(value)



# Copyright (c) 2017 Riverbed Technology, Inc.
#
# This software is licensed under the terms and conditions of the MIT License
# accompanying the software ("License").  This software is distributed "AS IS"
# as set forth in the License.
from django.conf import settings
from django.core.files.storage import FileSystemStorage

if not hasattr(settings, 'PCAP_STORE'):
    raise ValueError('Please set local_settings.PCAP_STORE to the proper '
                     'path for pcap storage')

pcap_store = FileSystemStorage(location='{0}/'.format(settings.PCAP_STORE),
                               base_url='{0}/'.format(settings.PCAP_STORE))



# Copyright (c) 2017 Riverbed Technology, Inc.
#
# This software is licensed under the terms and conditions of the MIT License
# accompanying the software ("License").  This software is distributed "AS IS"
# as set forth in the License.
import logging
import datetime
from os import stat, path
from django.contrib import messages
from django.core.urlresolvers import reverse
from django.forms.models import modelformset_factory
from django.http import HttpResponseRedirect, HttpResponse, \
    StreamingHttpResponse
from django.core.servers.basehttp import FileWrapper
from django.shortcuts import get_object_or_404

from rest_framework import generics, views
from rest_framework.renderers import TemplateHTMLRenderer, JSONRenderer
from rest_framework.permissions import IsAuthenticated, IsAdminUser
from rest_framework.response import Response

from steelscript.appfwk.apps.pcapmgr.models import pcap_store, PcapDataFile, \
    PcapFileField, WARN_PCAP
from steelscript.appfwk.apps.pcapmgr.forms import PcapFileForm, \
    PcapFileListForm
from steelscript.appfwk.apps.pcapmgr.serializers import PcapDataFileSerializer

logger = logging.getLogger(__name__)


class PcapFileDetail(views.APIView):
    """ Display File detail view """

    model = PcapDataFile
    serializer_class = PcapDataFileSerializer
    renderer_classes = (TemplateHTMLRenderer, JSONRenderer)
    permission_classes = (IsAuthenticated,)

    def get(self, request, pcapfile_id=None):
        if request.accepted_renderer.format == 'html':
            if pcapfile_id:
                datafile = get_object_or_404(PcapDataFile, pk=pcapfile_id)
                form = PcapFileForm(instance=datafile)
                f_name = path.basename(datafile.datafile.url)
            else:
                form = PcapFileForm()
                f_name = None
            return Response({'form': form, 'file_name': f_name},
                            template_name='pcapfile_detail.html')
        else:
            datafile = get_object_or_404(PcapDataFile, pk=pcapfile_id)
            serializer = PcapDataFileSerializer(instance=datafile)
            data = serializer.data
            return Response(data)

    def post(self, request, pcapfile_id=None):
        if pcapfile_id is not None:
            datafile = get_object_or_404(PcapDataFile, pk=pcapfile_id)
            form = PcapFileForm(request.DATA, instance=datafile)
        else:
            form = PcapFileForm(data=request.DATA,
                                files=request.FILES)

        if form.is_valid():
            form.save()
            return HttpResponseRedirect(reverse('pcapfile-list'))
        else:
            return Response({'form': form},
                            template_name='pcapfile_detail.html')

    def delete(self, request, pcapfile_id):
        datafile = get_object_or_404(PcapDataFile, pk=pcapfile_id)
        pcap_store.delete(datafile.datafile.name)
        datafile.delete()
        return HttpResponse(status=204)


class PcapFileList(generics.ListAPIView):
    model = PcapDataFile
    serializer_class = PcapDataFileSerializer
    renderer_classes = (TemplateHTMLRenderer, JSONRenderer)
    permission_classes = (IsAuthenticated,)

    supported_files = ', '.join(model.SUPPORTED_FILES.keys())

    def get(self, request, *args, **kwargs):
        queryset = PcapDataFile.objects.order_by('id')

        if request.accepted_renderer.format == 'html':
            df_form_set = modelformset_factory(PcapDataFile,
                                               form=PcapFileListForm,
                                               extra=0)
            formset = df_form_set(queryset=queryset)
            tabledata = queryset
            data = {'formset': formset, 'tabledata': tabledata,
                    'supported_files': self.supported_files,
                    'pcap_lib_warning': WARN_PCAP}
            return Response(data, template_name='pcapfile_list.html')

        serializer = PcapDataFileSerializer(instance=queryset)
        data = serializer.data
        return Response(data)


class PcapFSSync(views.APIView):
    permission_classes = (IsAdminUser,)

    def get(self, request):
        """
        Runs an operation to sync the file system and the DB.
        redirects to the list view.
        """

        added_files = 0
        removed_db = 0
        ignored_files = 0

        removed_storage_loc = False

        # First check that the directory is present. If not then
        # there can't be any files.
        if path.exists(pcap_store.location):
            fs_files = list()
            _, fs_raw_files = pcap_store.listdir(pcap_store.location)
            for f in fs_raw_files:
                t_name, t_code = PcapFileField.get_magic_type(
                    pcap_store.open(f)
                )
                # Only take files supported by the field and storage.
                # ignore the rest.
                if t_name:
                    fs_files.append(f)
                else:
                    ignored_files += 1
            db_files = PcapDataFile.objects.order_by('id')

            # first pass look in the db and delete any records that
            # don't have a file system object.
            for (fname, datafile) in [(path.basename(dbfile.datafile.name),
                                       dbfile) for dbfile in db_files]:
                if not fs_files.count(fname):
                    datafile.delete()
                    removed_db += 1

            # now look over the file system files to see if are are not
            # in the DB
            db_file_names = [path.basename(dbfile.datafile.name) for dbfile in
                             PcapDataFile.objects.order_by('id')]
            for fsfile in fs_files:
                if not db_file_names.count(fsfile):
                    f = pcap_store.path(fsfile)
                    t_name, t_sig = \
                        PcapFileField.get_magic_type(pcap_store.open(f))
                    new_db = PcapDataFile(description=fsfile,
                                          uploaded_at=(
                                              pcap_store.created_time(fsfile)),
                                          file_type=t_name,
                                          datafile=f,
                                          start_time=datetime.datetime.now(),
                                          end_time=datetime.datetime.now()
                                          )
                    new_db.save()
                    added_files += 1
        else:
            db_files = PcapDataFile.objects.order_by('id')
            if db_files:
                removed_storage_loc = True

        if removed_storage_loc:
            msg = ('Warning: PCAP records found in the database but the '
                   'underlying storage location is missing. This could be '
                   'caused by issues such as file system corruption. Please i'
                   'nvestigate.')
        else:
            msg = ('PCAP Manager Sync: {0} PCAP record(s) removed from '
                   'database. {1} new file object(s) added to database{2}')

            skipped = '.'
            if ignored_files:
                skipped = ' ({0} file(s) with invalid type skipped).'.format(
                    ignored_files
                )
            msg = msg.format(removed_db, added_files, skipped)

        messages.add_message(request._request, messages.INFO, msg)

        return HttpResponseRedirect(reverse('pcapfile-list'))


def pcap_download(request, pcap_name=None):
    if pcap_name is not None:
        file_path = pcap_store.path(pcap_name)
        chunk_size = 16384
        response = StreamingHttpResponse(FileWrapper(open(file_path,
                                                          'rb'),
                                         chunk_size))

        response['Content-Type'] = 'application/vnd.tcpdump.pcap'
        response['Content-Length'] = \
            stat(pcap_store.path(pcap_name)).st_size
        response['Content-Disposition'] = ('attachment; filename={0}'
                                           ''.format(pcap_name))
        return response
    else:
        return HttpResponseRedirect(reverse('pcapfile-list'))



# Copyright (c) 2015 Riverbed Technology, Inc.
#
# This software is licensed under the terms and conditions of the MIT License
# accompanying the software ("License").  This software is distributed "AS IS"
# as set forth in the License.

import logging

from collections import namedtuple

from elasticsearch_dsl import Search
from elasticsearch_dsl.connections import connections
from elasticsearch import helpers
from elasticsearch.connection.base import logger as elastic_logger

from steelscript.common.timeutils import datetime_to_seconds
from steelscript.appfwk.project.settings import ELASTICSEARCH_HOSTS

logger = logging.getLogger(__name__)

# Suppress debug logging in elasticsearch.connection.base module
elastic_logger.setLevel(logging.INFO)

MAX_NUMBER_DOCS = 10000

ColumnFilter = namedtuple('ColumnFilter', ['query_type', 'query'])


class ElasticSearch(object):

    def __init__(self):
        self.client = connections.create_connection(hosts=ELASTICSEARCH_HOSTS)

    def write(self, index, doctype, data_frame, timecol):

        df = data_frame.fillna('')

        actions = [{"_index": index,
                    "_type": doctype,
                    "_id": datetime_to_seconds(df.iloc[i][timecol]),
                    "_source": df.iloc[i].to_dict()}
                   for i in xrange(len(df))]

        logger.debug("Writing %s records from %s to %s into db. Index: %s, "
                     "doc_type: %s."
                     % (len(df), df[timecol].min(), df[timecol].max(),
                        index, doctype))

        written, errors = helpers.bulk(self.client, actions=actions,
                                       stats_only=True)
        logger.debug("Successfully wrote %s records, %s errors." % (written,
                                                                    errors))
        return

    def search(self, index, doc_type, col_filters=None):

        logger.debug("Searching index %s for doc_type %s and col_filters %s"
                     % (index, doc_type, col_filters))

        s = Search(using=self.client, index=index, doc_type=doc_type)
        if col_filters:
            for col_filter in col_filters:
                if isinstance(col_filter, ColumnFilter):
                    s = s.filter(col_filter.query_type, **col_filter.query)
                else:
                    raise ValueError('Column Filter is not an instance of'
                                     ' ColumnFilter class')
        s = s.params(size=MAX_NUMBER_DOCS)

        results = s.execute()

        logger.debug("Search returned %s records from elasticsearch."
                     % len(results))
        return [res.to_dict() for res in results]



# Copyright (c) 2017 Riverbed Technology, Inc.
#
# This software is licensed under the terms and conditions of the MIT License
# accompanying the software ("License").  This software is distributed "AS IS"
# as set forth in the License.

from rest_framework.exceptions import APIException


class NotFoundError(APIException):
    """ No object found for the handle. """
    status_code = 404
    default_detail = 'Data not found.'


class InvalidRequest(APIException):
    """Invalid query parameter error. """
    status_code = 400
    default_detail = "Invalid request."



# Copyright (c) 2015 Riverbed Technology, Inc.
#
# This software is licensed under the terms and conditions of the MIT License
# accompanying the software ("License").  This software is distributed "AS IS"
# as set forth in the License.

import sys
import csv

from django.core.management.base import BaseCommand, CommandError
from django.db import transaction

from steelscript.common.datautils import Formatter
from steelscript.appfwk.apps.devices.models import Device

REQUIRED_DEVICE_ATTRS = ['host', 'module', 'name', 'username', 'password']
DEVICE_ATTRS = REQUIRED_DEVICE_ATTRS + ['enabled', 'port']


class Command(BaseCommand):
    args = ''
    help = 'Add, modify, delete and display device objects'

    def add_arguments(self, parser):
        parser.add_argument('--show',
                            action='store_true',
                            dest='show_device',
                            default=False,
                            help=('Show one or more matching device(s). '
                                  'For example:\n'
                                  './manage.py device --show --enabled'))

        parser.add_argument('--add',
                            action='store_true',
                            dest='add_device',
                            default=False,
                            help=('Add one device to the database. '
                                  'For example:\n'
                                  './manage.py device --add -H host -M module '
                                  '-N name -p port -U user -P password '
                                  '-E True'))

        parser.add_argument('--edit',
                            action='store_true',
                            dest='edit_device',
                            default=False,
                            help=('Edit one device identified by id. '
                                  'For example:\n'
                                  './manage.py device --edit -I 1 '
                                  '-N new_name'))

        parser.add_argument('--batch-file',
                            dest='batch_file',
                            help=("Batch file with info of multiple "
                                  "devices in the format of "
                                  "<name>,<module>,<hostname>,<port>,"
                                  "<username>,<password>,<AuthMethod>"
                                  "<access_code>,<tag1;tag2;...>"
                                  "for AuthMethod, 1 refers to "
                                  "username/password; 2 refers to OAuth2."
                                  ))

        parser.add_argument('--enabled', action='store_true', dest='enabled',
                            help='Set the device as enabled')
        parser.add_argument('--no-enabled', action='store_false',
                            dest='enabled', help='Set the device as disabled')

        parser.add_argument('-H', '--host', help='Hostname or IP address')
        parser.add_argument('-M', '--module',
                            help='Module name for the device')
        parser.add_argument('-N', '--name', help='Name of the device')
        parser.add_argument('-p', '--port',
                            help='Port of the device to connect')
        parser.add_argument('-U', '--username', help='Username for the device')
        parser.add_argument('-P', '--password', help='Password for the device')
        parser.add_argument('-I', '--id', help='ID of a device to be edited')

    def handle(self, *args, **options):
        """ Main command handler. """
        if options['show_device']:
            dev = {}
            for opt in DEVICE_ATTRS + ['id']:
                if options[opt] is not None:
                    dev[opt] = options[opt]

            output = []
            for d in Device.objects.filter(**dev):
                output.append([d.id, d.name, d.module, d.host, d.port,
                               d.username, d.enabled])
            Formatter.print_table(output, ['ID', 'Name', 'Module', 'Host',
                                           'Port', 'User Name', 'Enabled'])
            if not output:
                self.stdout.write("No device found")

        elif options['add_device']:
            dev = {}
            for opt in REQUIRED_DEVICE_ATTRS:
                if options[opt] is None:
                    self.stdout.write("Option '%s' is required for a device" %
                                      opt)
                    sys.exit(1)
                dev[opt] = options[opt]

            if options['port']:
                if options['port'].isdigit():
                    dev['port'] = int(options['port'])
                else:
                    self.stdout.write("Option port '%s' is not a positive "
                                      "integer" % options['port'])
                    sys.exit(1)

            if options['enabled'] is not None:
                # --enabled or --no-enabled is set
                dev['enabled'] = options['enabled']

            dev_obj = Device(**dev)
            dev_obj.save()
            self.stdout.write('Device added.')

        elif options['edit_device']:

            if not options['id']:
                self.stdout.write('Option ID is required to edit a device')
                sys.exit(1)

            elif not options['id'].isdigit():
                self.stdout.write("Option id '%s' is not a positive integer" %
                                  options['id'])
                sys.exit(1)

            dev_list = Device.objects.filter(id=int(options['id']))
            if not dev_list:
                self.stdout.write("Option id '%s' does not match a device" %
                                  options['id'])
                sys.exit(1)

            if options['port']:
                if options['port'].isdigit():
                    options['port'] = int(options['port'])
                else:
                    self.stdout.write("Option port '%s' is not a positive "
                                      "integer" % options['port'])
                    sys.exit(1)

            dev_obj = dev_list[0]
            change = False
            for attr in DEVICE_ATTRS:
                if options[attr] is not None:
                    change = True
                    setattr(dev_obj, attr, options[attr])

            if change:
                dev_obj.save()
                self.stdout.write("Device '%s' modified" % options['id'])
            else:
                self.stdout.write("Device '%s' unchanged" % options['id'])

        elif options['batch_file']:
            file_name = options['batch_file']
            default_header = ['name', 'module', 'host', 'port',
                              'username', 'password', 'auth',
                              'access_code', 'tags']
            with open(file_name, 'r') as f:
                reader = csv.reader(f)
                devs = []
                add, update = 0, 0
                for i, row in enumerate(reader):
                    row = map(str.strip, row)
                    if i == 0:
                        if set(map(str.lower, row)) == set(default_header):
                            header = map(str.lower, row)
                            continue
                        else:
                            header = default_header

                    if not row or not row[0]:
                        continue

                    if len(row) < len(header):
                        msg = ('Line {0} only has {1} fields. '
                               '{2} fields are required.'
                               .format(i+1, len(row), len(header)))
                        raise CommandError(msg)

                    kwargs = dict(zip(header, row))

                    if not kwargs['host']:
                        msg = 'Host is empty string in line {0}.'.format(i+1)
                        raise CommandError(msg)

                    if not kwargs['port'].isdigit():
                        msg = ("Port should be integer instead of '{0}'"
                               " in line {1}".format(kwargs['port'], i+1))
                        raise CommandError(msg)

                    kwargs['port'] = int(kwargs['port'])
                    kwargs['enabled'] = True
                    kwargs['tags'] = ','.join(map(str.strip,
                                                  kwargs['tags'].split(';')))

                    # Check for device with same host and port
                    res = Device.objects.filter(host=kwargs['host'],
                                                port=kwargs['port'])
                    if res:
                        dev = res[0]
                        for k, v in kwargs.iteritems():
                            setattr(dev, k, v)
                        update += 1
                    else:
                        dev = Device(**kwargs)
                        add += 1

                    devs.append(dev)

                with transaction.atomic():
                    for dev in devs:
                        dev.save()

                self.stdout.write('{add} devices added. '
                                  '{update} existing devices refreshed.'
                                  .format(add=add, update=update))



# Copyright (c) 2015 Riverbed Technology, Inc.
#
# This software is licensed under the terms and conditions of the MIT License
# accompanying the software ("License").  This software is distributed "AS IS"
# as set forth in the License.

import copy
import pandas
import logging
import hashlib

from functools import partial

from steelscript.appfwk.apps.jobs.models import Job
from steelscript.appfwk.apps.datasource.models import Table
from steelscript.appfwk.apps.datasource.modules.analysis import AnalysisTable,\
    AnalysisQuery
from steelscript.appfwk.apps.jobs import QueryComplete, QueryContinue
from steelscript.appfwk.apps.db import storage, ColumnFilter, \
    ExistingIntervals
from steelscript.common.interval import Interval, IntervalList
from steelscript.appfwk.apps.datasource.models import Column
from steelscript.common.timeutils import round_time, timedelta_total_seconds,\
    force_to_utc, TimeParser

logger = logging.getLogger(__name__)

TIME_FIELDS = ['_orig_duration', '_orig_endtime', '_orig_starttime',
               'duration', 'endtime', 'starttime']

tp = TimeParser()


def make_index(s):
    return 'appfwk-{0}'.format(s)


class TimeInterval(Interval):
    def __init__(self, start, end):
        self.start = force_to_utc(start)
        self.end = force_to_utc(end)

    def localize_tz(self, tzinfo):
        self.start = self.start.astimezone(tzinfo)
        self.end = self.end.astimezone(tzinfo)


class TimeSeriesTable(AnalysisTable):

    class Meta:
        proxy = True

    _query_class = 'TimeSeriesQuery'

    TABLE_OPTIONS = {'max_length_per_call': 3600,
                     'max_number_of_calls': 2}


class TimeSeriesQuery(AnalysisQuery):

    def __init__(self, *args, **kwargs):
        super(AnalysisQuery, self).__init__(*args, **kwargs)

        self.ds_table = Table.from_ref(self.table.options.related_tables['ds'])

        self.time_col = [col.name for col in self.ds_table.get_columns()
                         if col.datatype == Column.DATATYPE_TIME][0]

        starttime, endtime, self.resolution = self._round_times()

        self.query_interval = TimeInterval(starttime, endtime)

        self.handle, self.no_time_criteria = self._calc_handle()

    def _calc_handle(self):
        """ Compute the handle corresponding to the table object
        and the criteria fields, excluding fields such as 'endtime',
        'starttime' and 'duration'.

        :return: the hash hex string.
        """
        h = hashlib.md5()
        h.update(str(self.table.id))

        h.update('.'.join([c.name for c in self.ds_table.get_columns()]))

        criteria = copy.copy(self.job.criteria)

        for k, v in self.job.criteria.iteritems():
            if k in TIME_FIELDS:
                # We do not want time related fields included in the criteria
                criteria.pop(k, None)
                continue
            h.update('%s:%s' % (k, v))

        return h.hexdigest(), criteria

    def _check_intervals(self, intervals):

        max_len = self.table.options.max_length_per_call * self.resolution
        max_calls = self.table.options.max_number_of_calls

        ret = IntervalList([])

        err_msg = ("Could not fulfill the query for %s. "
                   "Max length for each query is %s. "
                   "Max number of queries is %s. "
                   % (intervals, max_len, max_calls))

        calls_left = max_calls

        for interval in intervals:
            if calls_left <= 0 or \
                    interval.size > max_len * calls_left:
                raise ValueError(err_msg)
            if interval.size <= max_len:
                ret.append(interval)
                calls_left -= 1
            else:  # interval.size <= max_len * calls_left:
                while interval.size > max_len:
                    one_int = TimeInterval(interval.start,
                                           interval.start + max_len)
                    interval = (interval - one_int)[0]
                    calls_left -= 1
                    ret.append(one_int)

                # Add the remaining interval
                ret.append(interval)
                calls_left -= 1

        return ret

    def _round_times(self):
        """Round the start/end time in criteria to reflect what data will be
        returned from data source and saved in db (based on investigation of
        NetProfiler/NetShark query results).

        Start/End time needs to round down to closest time in integer
        resolutions. However, if either starttime or endtime in criteria is
        already in integer resolutions, it should remain the same.

        :return: start time, end time, resolution
        """

        self.job.criteria.compute_times()

        resolution = getattr(self.job.criteria, 'resolution', None)
        if not resolution:
            raise AttributeError("The data source table '%s' needs to have "
                                 "'resolution' field." % self.ds_table.name)

        resolution_sec = timedelta_total_seconds(self.job.criteria.resolution)

        starttime = round_time(self.job.criteria.starttime,
                               round_to=resolution_sec,
                               trim=True)

        endtime = round_time(self.job.criteria.endtime,
                             round_to=resolution_sec,
                             trim=True)

        return starttime, endtime, resolution

    def query(self, starttime, endtime):
        col_filters = [ColumnFilter(
                       query_type='range',
                       query={self.time_col: {'gte': starttime,
                                              'lte': endtime}
                              })]

        # Obtain result from storage as a list of dicts,
        # each dict represents one record/doc
        res = storage.search(index=make_index(self.ds_table.namespace),
                             doc_type=self.handle,
                             col_filters=col_filters)

        # The time field is a string formatted as "YYYY/MM/DDTHH:MM:SS"
        # Need to convert it to datetime type
        if not res:
            return None

        make_ts = partial(pandas.Timestamp, tz='UTC')
        df = pandas.DataFrame(res)
        df[self.time_col] = df[self.time_col].map(make_ts)

        return df.sort(self.time_col)

    def _converge_adjacent(self, intervals):

        itvs = []
        current_interval = None
        for interval in intervals:
            if current_interval is None:
                current_interval = interval
            elif current_interval.end + self.resolution == interval.start:
                current_interval = TimeInterval(current_interval.start,
                                                interval.end)
            else:
                itvs.append(current_interval)
                current_interval = interval
        itvs.append(current_interval)
        return IntervalList(itvs)

    def analyze(self, jobs=None):

        filtered_list = ExistingIntervals.objects.filter(
            table_handle=self.handle)

        existing_intervals = None

        if filtered_list:
            existing_intervals = filtered_list[0].intervals

            if self.query_interval in existing_intervals:
                # Search DB for the queried data
                return QueryComplete(self.query(self.query_interval.start,
                                                self.query_interval.end))

        intervals_to_call = self._check_intervals(
            self.query_interval - existing_intervals)

        dep_jobs = {}
        for interval in intervals_to_call:
            criteria = copy.copy(self.job.criteria)
            # Use the two time related fields
            criteria.starttime = interval.start
            criteria.endtime = interval.end
            job = Job.create(table=self.ds_table, criteria=criteria,
                             update_progress=False, parent=self.job)
            dep_jobs[job.id] = job

        return QueryContinue(self.collect, jobs=dep_jobs)

    def collect(self, jobs=None):
        dfs_from_jobs, dfs_from_db = [], []

        objs = ExistingIntervals.objects.filter(table_handle=self.handle)

        if objs:
            obj = objs[0]

            # Query for the intervals already in db
            itvs_in_db = self.query_interval.intersection(obj.intervals)
            dfs_from_db = [self.query(itv.start, itv.end)
                           for itv in itvs_in_db]
        else:
            obj = ExistingIntervals(namespace=self.ds_table.namespace,
                                    sourcefile=self.ds_table.sourcefile,
                                    table=self.ds_table.name,
                                    criteria=self.no_time_criteria,
                                    table_handle=self.handle,
                                    intervals=IntervalList([]))

        for job_id, job in jobs.iteritems():
            df = job.data()
            if df is None:
                continue
            dfs_from_jobs.append(df)

            obj.intervals += \
                TimeInterval(df[self.time_col].min().to_datetime(),
                             df[self.time_col].max().to_datetime())

        if not dfs_from_jobs:
            return QueryComplete(None)

        storage.write(index=make_index(self.ds_table.namespace),
                      doctype=self.handle,
                      data_frame=pandas.concat(dfs_from_jobs,
                                               ignore_index=True),
                      timecol=self.time_col)

        obj.intervals = self._converge_adjacent(obj.intervals)

        obj.tzinfo = self.job.criteria.starttime.tzinfo

        # Only update existing intervals if writing to db succeeds
        obj.save()

        # Immediately reading from db after writing results can result in
        # non-correct data, thus stitching the data frames together
        total_df = pandas.concat(dfs_from_db + dfs_from_jobs,
                                 ignore_index=True)
        return QueryComplete(total_df.sort(self.time_col).drop_duplicates())



# Copyright (c) 2017 Riverbed Technology, Inc.
#
# This software is licensed under the terms and conditions of the MIT License
# accompanying the software ("License").  This software is distributed "AS IS"
# as set forth in the License.


from django.contrib import admin
from steelscript.appfwk.apps.pcapmgr.models import PcapDataFile


class PcapFileAdmin(admin.ModelAdmin):
    list_display = ('datafile', 'description', 'uploaded_at',
                    'file_type', 'start_time', 'end_time',
                    'pkt_count', 'packet_bytes', )


admin.site.register(PcapDataFile, PcapFileAdmin)



# Copyright (c) 2015 Riverbed Technology, Inc.
#
# This software is licensed under the terms and conditions of the MIT License
# accompanying the software ("License").  This software is distributed "AS IS"
# as set forth in the License.

import logging

from django.http import JsonResponse
from django.core.exceptions import ObjectDoesNotExist

from rest_framework import views

from steelscript.appfwk.apps.db import storage, ColumnFilter
from steelscript.common.timeutils import sec_string_to_datetime
from steelscript.appfwk.apps.db.models import ExistingIntervals
from steelscript.appfwk.apps.datasource.modules.ts_table import make_index
from steelscript.appfwk.apps.datasource.models import Table
from steelscript.appfwk.apps.datasource.models import Column
from steelscript.appfwk.apps.db.exceptions import NotFoundError, \
    InvalidRequest

logger = logging.getLogger(__name__)


class Records(views.APIView):

    def get(self, request):
        """ Retrieve records from time series data storage.

        The URL is formatted as '/db/records?handle=**&start=**&end=**'.
        Within the URL, required parameters include 'handle' and 'start'.
        Optional parameter is 'end'. Values for 'start' and 'end' in the
        URL should be epoch seconds. JSON results returned looks like:

        [{"avg_bytes": 1617806.0,
          "time": "2017-03-24T18:14:00+00:00"},
          ...
        ]
        """

        request_data = request.GET.dict()

        keys = ['handle', 'start']
        for k in keys:
            if k not in request_data:
                msg = "Missing parameter '{}' in url".format(k)
                raise InvalidRequest(msg)

        handle = request_data['handle']
        try:
            obj = ExistingIntervals.objects.get(table_handle=handle)
        except ObjectDoesNotExist:
            msg = "Handle '{}' does not exist.".format(handle)
            raise NotFoundError(msg)

        tr = {}
        tr['gte'] = sec_string_to_datetime(int(request_data['start']))
        if 'end' in request_data:
            tr['lte'] = sec_string_to_datetime(int(request_data['end']))

        # Getting the time column name
        table = Table.from_ref(dict(sourcefile=obj.sourcefile,
                                    namespace=obj.namespace,
                                    name=obj.table))

        timecols = [c for c in table.get_columns(iskey=True)
                    if c.datatype == Column.DATATYPE_TIME]

        time_col_name = timecols[0].name

        col_filters = [ColumnFilter(
                       query_type='range',
                       query={time_col_name: tr})]

        records = storage.search(index=make_index(obj.namespace),
                                 doc_type=handle,
                                 col_filters=col_filters)

        return JsonResponse(records, safe=False)


class Handles(views.APIView):

    def get(self, request):
        """ Retrieve time series data handles.

        The URL is formated as '/db/handles/?namespace=**&table=**'.
        Both 'namespace' and 'table' parameters are optional. 'namespace'
        refers to the plugin name while 'table' refers to the name of
        the timeseries source table generating the records.

        JSON results returned look like:
        [{"intervals": "[(2017-03-29 16:16:00+00:00,
                          2017-03-29 17:15:00+00:00),
                        (2017-03-29 18:16:00+00:00,
                         2017-03-29 19:13:00+00:00),
                        ...
                       ]",
          "handle": "d021d2d84a6c0367d9bcfee9f1ac1912",
          "criteria": "{u'netprofiler_filterexpr': u'',
                        'debug': False,
                        u'resolution': datetime.timedelta(0, 60),
                        u'netprofiler_device': u'1',
                        'ignore_cache': False}",
          "table": "ts-overall-db",
          "sourcefile": "reports.netprofiler.netprofiler_db",
          "namespace": "netprofiler"},
          ...
        ]
        """
        request_data = request.GET.dict()

        keys = ['namespace', 'table']
        for k, v in request_data.iteritems():
            if k not in keys:
                msg = "'{}' is not valid to query handles." .format(k)
                raise InvalidRequest(msg)

        res = []
        for obj in ExistingIntervals.objects.filter(**request_data):
            res.append(dict(handle=obj.table_handle,
                            criteria=str(obj.criteria),
                            namespace=obj.namespace,
                            sourcefile=obj.sourcefile,
                            table=obj.table,
                            intervals=str(obj.intervals)))
        return JsonResponse(res, safe=False)



# Copyright (c) 2015 Riverbed Technology, Inc.
#
# This software is licensed under the terms and conditions of the MIT License
# accompanying the software ("License").  This software is distributed "AS IS"
# as set forth in the License.


import os
import json
import logging
from cStringIO import StringIO
from tagging_autocomplete.models import TagAutocompleteField
from django.utils.translation import ugettext_lazy as _
from django.core.exceptions import ValidationError

from django.db import models
from django.core import management
from django.conf import settings

from steelscript.common import Auth

logger = logging.getLogger(__name__)


def validate_port(value):
    if value < 1 or value > 65535:
        raise ValidationError(
            _('%(value)s is not a valid port number, '
              'which should be between 1 and 65535'),
            params={'value': value},
        )


def create_device_fixture(strip_passwords=True):
    """ Dump devices to JSON file, optionally stripping passwords.

    If password is stripped, the device is disabled.
    """
    buf = StringIO()
    management.call_command('dumpdata', 'devices', stdout=buf)
    buf.seek(0)
    devices = list()
    for d in json.load(buf):
        if strip_passwords:
            del d['fields']['password']
            del d['fields']['access_code']
            d['fields']['enabled'] = False
        devices.append(d)

    buf.close()

    fname = os.path.join(settings.INITIAL_DATA, 'initial_devices.json')
    with open(fname, 'w') as f:
        f.write(json.dumps(devices, indent=2))

    logger.debug('Wrote %d devices to fixture file %s' % (len(devices), fname))


class DevManager(models.Manager):

    def filter_by_tag(self, tag, **kwargs):

        dev_ids = [dev.id for dev in Device.objects.filter(**kwargs)
                   if tag in dev.tags_as_list()]

        return Device.objects.filter(pk__in=dev_ids)


class Device(models.Model):
    """ Records for devices referenced in report configuration pages.

        Actual instantiations of Device objects handled through DeviceManager
        class in devicemanager.py module.
    """
    name = models.CharField(max_length=200)
    module = models.CharField(max_length=200)
    host = models.CharField(max_length=200)
    port = models.IntegerField(default=443, validators=[validate_port])
    username = models.CharField(max_length=100, blank=True)
    password = models.CharField(max_length=100, blank=True)
    tags = TagAutocompleteField(blank=True)

    objects = DevManager()

    auth = models.IntegerField(
        default=Auth.NONE,
        choices=((Auth.NONE, 'None'),
                 (Auth.BASIC, 'Basic'),
                 (Auth.OAUTH, 'OAuth2'))
    )

    access_code = models.TextField(blank=True)

    # only enabled devices will require field validation
    enabled = models.BooleanField(default=True)

    class Meta:
        unique_together = (("host", "port"))

    def __unicode__(self):
        return '%s (%s:%s)' % (self.name, self.host, self.port)

    def save(self, *args, **kwargs):
        super(Device, self).save(*args, **kwargs)
        create_device_fixture(settings.APPFWK_STRIP_DEVICE_PASSWORDS)

    def tags_as_list(self):
        return [tag.strip() for tag in self.tags.split(',') if tag.strip()]
